
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: julien de Rosny (c) Institut Langevin/ESPCI/CNRS
"""
from numpy import *
from scipy.optimize import basinhopping,minimize,least_squares,minimize_scalar
from matplotlib.pyplot import *
from mutual import *
from numpy import *
from numpy.linalg import *
from numpy.random import *
from utilsDraw import PlotWires
import sys

# Standarised capa values
stdCapas = array([1,1.1,1.2,1.3,1.5,1.6,1.8,2.0,2.2,2.4,2.7,3.0,3.3,3.6,3.9,4.3,4.7,5.1,5.6,6.2,6.8,7.5,8.2,9.1])
# Void impedance
eta = 376.730313668
close('all')
coefk = 0
N = 11 # Number of Legs
thetaApperture = 45.0/180.0*pi*2
a1 = 0.123e-2 #inner diammeter of legs
a2 = 0.198e-2 #outer diammeter of legs
widthLegMax = 0.01*2 # Width of the leg
widthLegMin = 0.005 # Width of the leg
widthRingMax = 0.02 # Width of the leg
widthRingMin = 0.005 # Width of the leg
lengthLeg = 24e-2 # Length of legs
R0 = 15.5e-2   # Radius of shield
Rleg = 13e-2 # Radius of circle where the leg are positionned
Rbody = 0.09 # Radius of imaging
omega = 300e6*2*pi # Working frequency
deltaThetaLengthRing = 0.005/Rleg
k0 = omega/3e8 #Wave nuumber
widthMin = 2e-3 #Width leg min
widthMax =  2*pi*Rleg/N/2.0 #Width leg max
Cmin = 1e-12 #Min capacity
flagSym = True
CrTest = 2.17307e-12
ZminCLeg = -1/Cmin/omega # Minmum value of serial impedance value on Leg
ZminCRing = -1/Cmin/omega # Minmum value of serial impedance value on Leg



        
# Definition of a function to transform a 2M vector into a vector of M+1 value,
# is just the complex sum of the M first elements
# This is done because we know that it is not possible to generate constant c
# current with a metacage
def Complexi(x):
    v=x[0::2]+1j*x[1::2]
    somme=np.sum(v)
    return np.append(v,-somme)

theta = np.linspace(thetaApperture/2,2*np.pi-thetaApperture/2,N)
#theta = np.linspace(0,2*np.pi-2*pi/N,N)

# write geometrical values
with open('geom.inc','wt') as f:
    f.write('RbodyFile=%e;\n'%Rbody)
    f.write('RairFile=%e;\n'%R0)
    f.write('RbirdFile=%e;\n'%Rleg)
    f.write('NthetaFile=%d;\n'%len(theta))
    for  it,t in enumerate(theta):
        f.write('thetaFile~{%d}=%e;\n'%(it,t))


deltaTheta = theta[1] - theta[0]
# xy are the positions of the legs (required to compute mutual impedance) 
xyz  = Rleg *  column_stack((cos(theta),sin(theta),zeros((len(theta))))).transpose()
# Radius of the circle of the image of the leg circle with respect to shield
Rp = R0**2/Rleg
# xyp positions of the leg images
xyzp  = Rp *  column_stack((cos(theta),sin(theta),zeros((len(theta))))).transpose()
#
# xy are the positions of the legs (required to compute mutual impedance) 
ringSegments  = [\
                 array([[Rleg*cos(theta[i]+deltaThetaLengthRing)           ,Rleg*sin(theta[i]+deltaThetaLengthRing)           ,0], \
                        [Rleg*cos(theta[(i+1)%len(theta)]-deltaThetaLengthRing),Rleg*sin(theta[(i+1)%len(theta)]-deltaThetaLengthRing),0]]) \
                 for i in range(len(theta))\
                ]

ringLengths = [linalg.norm(c[0,:]-c[1,:]) for c in ringSegments]
                 
legSegments = [ array([[Rleg*cos(t),Rleg*sin(t),0],[Rleg*cos(t),Rleg*sin(t),lengthLeg]])  for t in theta ]


# xy are the positions of the legs (required to compute mutual impedance) 
                 
legSegmentsImage = [ array([[Rp*cos(t),Rp*sin(t),0],[Rp*cos(t),Rp*sin(t),lengthLeg]])  for t in theta ]


imMinLeg  = [selfInducRect(lengthLeg,widthRingMax)*1e-7*omega]*ones(N)
imMaxLeg  = [selfInducRect(lengthLeg,widthRingMin)*1e-7*omega]*ones(N)


imMinRing  = array([selfInducRect(l,widthRingMax)*1e-7*omega for l in ringLengths ])
imMaxRing  = array([selfInducRect(l,widthRingMin)*1e-7*omega for l in ringLengths])

if 0:
    figure()
    plt3 = PlotWires()
    plt3.plot(ringSegments)
    plt3.plot([s + array([[0,0,lengthLeg],[0,0,lengthLeg]]) for s in ringSegments])
    plt3.plot(legSegments,'black')
    plt3.plot(legSegmentsImage,'grey')
    axis('off')
#

# Calculus of the self and mutual inductance
def ComputeRoverHalf(k0):
    coef = eta/3/2/pi*k0**2
    return((lengthLeg**2*eye(N)/2,diag(ringLengths)**2))
    
def ComputeL(k0,widthLeg=None,widthRing=None):
    LLleg =  zeros((N,N),dtype=complex)
    for i in range(N):
        for j in range(N):
            dima =  norm([xyz[:,i]-xyzp[:,j]]) 
            if i!=j:
                d =  norm([xyz[:,i]-xyz[:,j]]) 
                LLleg[i,j] = Lnm(lengthLeg,d,k0) - Lnm(lengthLeg,dima,k0)
            else:
                if widthLeg is not None:
                    LLleg[i,j] += selfInducRect(lengthLeg,widthLeg[i])
                LLleg[i,j] -=  Lnm(lengthLeg,dima,k0)#selfInducTub(lengthLeg,a1,a2)-Lnm(lengthLeg,dima)
            
    LLring = zeros((N,N),dtype=complex)
    for i in range(N):
        for j in range(N):
            #print(i,j)
    
            if i!=j:
                
    #            LLring[i,j] = MNonPara( column_stack((xyz[:,i],xyz[:,(i+1)%len(theta)])).transpose(), \
    #                                    column_stack((xyz[:,j],xyz[:,(j+1)%len(theta)])).transpose())
                LLring[i,j] = MNonPara( ringSegments[i],ringSegments[j],k0)
                if 0:
                    clf()
                    plot([xyz[0,i],xy[0,(i+1)%len(theta)]],[xyz[1,i],xyz[1,(i+1)%len(theta)]])
                    plot([xyz[0,j],xy[0,(j+1)%len(theta)]],[xyz[1,j],xyz[1,(j+1)%len(theta)]])
                    pause(.1)
            else:
                if widthRing is not None:
                    l = norm(xyz[:,i]-xyz[:,(i+1)%len(theta)])
                    LLring[i,j] = selfInducRect(l,widthRing[i])
                #LLring[i,j] -= MNonPara(column_stack((xyz[:,i],xyz[:,(i+1)%len(theta)])).transpose(), \
                #                     column_stack((xyzp[:,j],xyzp[:,(j+1)%len(theta)])).transpose())
    return(LLleg,LLring)
    
# Calculus of the self and mutual inductance. Tale into account
def ComputeLHalf(k0,widthLeg=None,widthRing=None):
    LLleg,LLring = ComputeL(k0,widthLeg,widthRing)
    return(LLleg/2.0,LLring)

    

#Compute mutual inductance (not self)
LLeg0,LRing0 = ComputeLHalf(k0*coefk,None,None)
#Compute resitive impedance due to radiative losses
RselfLegHalf, RselfRing = ComputeRoverHalf(k0)

# Current distribution in case of a BIRDCAGE
Il0 = real(exp(1j*theta))
Il0 = Il0 - mean(Il0)
#NEW PART


# We define somme simple matricial operators the shift a vector circurlarly such as n->n-1 (DOWN) n->n+1 (UP) and compute the difference between n+1 and n
# 
DOWN = zeros((N,N))
for i in range(N):
    DOWN[i,(i+1)%N] = 1
UP = inv(DOWN)
DELTA = UP-eye(N)

imMaxLeg = array([selfInducRect(lengthLeg,widthLegMin)*1e-7*omega/2.0]*N)
imMinLeg = array([selfInducRect(lengthLeg,widthLegMax)*1e-7*omega/2.0]*N)
imMaxRing =  array([selfInducRect(L,widthRingMin)*1e-7*omega for L in ringLengths])
imMinRing =  array([selfInducRect(L,widthRingMax)*1e-7*omega for L in ringLengths])

rotR = lambda x:array([x[(i-1)%len(x)] for i in range(len(x))])

dirac = zeros(N)
dirac[0] = 1
ZtildeLeg=fft.fft(LLeg0[:,0]*omega*1e-7 + imMaxLeg[0]*dirac)

ZtildeRing=fft.fft(LRing0[:,0]*omega*1e-7)+imMaxRing[0]

zetaR= -(ZtildeRing+4*sin(pi/N*arange(N))**2*ZtildeLeg)

CR=-1/np.real(zetaR)/omega    
CR[1]

Il1 = Il0#+1j*1e-10

imZl =  -0.0/omega/1e-12*ones(N) + imMaxLeg
imZr =  -1.0/omega/CR[1]*ones(N) + imMaxRing

Gamma = lambda ZL,ZR:(eye(N)-DOWN)@inv(ZR)@(eye(N)-UP) + inv(ZL)


def INode(imZl,imZr):
    # Leg impedance
    zetaL = LLeg0*1e-7*1j*omega  + RselfLegHalf + 1j*diag(imZl)
    zetaR = LRing0*1e-7*1j*omega + RselfRing + 1j*diag(imZr)
    V = zetaL@Il1
    IR = inv(zetaR)@(eye(N)-UP)@V
    I = (eye(N)-DOWN)@IR + Il1
    G = Gamma(zetaL,zetaR)
    return(I,V,G,zetaL,zetaR)

i,v,g,zetaL,zetaR = INode(imZl,imZr)
norm(i)

def INodeX(x):
    if flagSym:
        dMimZl,dMimZr = symZ(x) 
    else:
        dMimZl,dMimZr = x[:N ],x[N:]

    M,v,g,zetaL,zetaR = INode(dMimZl,dMimZr)
    return(M,v,g)

#ZselfLegImg = imMinLeg
#ZselfRingImg = imMinRing
#ZLegTot = lambda imZl,omega : LLeg0*1e-7*1j*omega + 1j*diag(imZl+ZselfLegImg)
#ZRingTot = lambda imZr,omega :  LRing0*1e-7*1j*omega + 1j*diag(imZr+ZselfRingImg)
# function to optimze (to get an eigenmode, we need the minimum input current)
def toOptimize(x):
    if flagSym:
        dMimZl,dMimZr = symZ(x) 
    else:
        dMimZl,dMimZr = x[:N ],x[N:]

    M,v,g = INodeX(x)
    y = norm(M)**2
    #print(y)
    return(y)
    
# function to generate symetric values
# 
def symZ(x):

#
    if True:
        l  = x[0:N//2]
        l1 = x[N//2]
        r  = x[N//2+1:N]
        r1 = x[N]
        imZl = concatenate((l,[l1],l[::-1]))
        imZr = concatenate((r,r[::-1],[r1]))
    else:
        imZl=x[0::2]
        imZr=x[1::2]
        if (len(x) % 2) == 0:
            imZl = concatenate((imZl,imZl[-1:0:-1]))
            imZr = concatenate((imZr,imZr[-2::-1]))
        else:            
            imZl = concatenate((imZl,imZl[-2:0:-1]))
            imZr = concatenate((imZr,imZr[-1::-1]))
    return(imZl,imZr)

# Compute the closest capacitance
def Capas(imZT,imMin):
    Clump =zeros(len(imZT))
    imZ =zeros(len(imZT))
    ZselfImg =zeros(len(imZT))
    for i in range(len(imZT)):
        if imZT[i] < imMin[i]:
            ZselfImg[i] = imZT[i]
            Clump[i] = 0
            while ZselfImg[i] < imMin[i]:
                delta = ZselfImg[i] - imMin[i]
                C = (-1/delta/omega)/1e-12 # C in  nf
                tenth = int(log10(C))
                Cref = stdCapas*10.0**tenth
                j = 0
                for Cr in Cref:
                    if Cr > C:
                        break
                Clump[i] += Cr 
                imZ[i] =  - 1 / omega / (Clump[i]*1e-12)
                ZselfImg[i] = imZT[i] - imZ[i]
        else:
            ZselfImg[i] = imZT[i]
            imZ[i] = 0
            Clump[i] = -float('inf')
    return(Clump,imZ,ZselfImg)

def Capas2(imZT,imMin,imMax):
    Clump =zeros(len(imZT))
    imZ =zeros(len(imZT))
    ZselfImg =zeros(len(imZT))
    Call = []
    for i in range(len(imZT)):
        if imZT[i] < imMin[i]:
            Cmin = 1/(imMax[i] - imZT[i])/omega
            Cmax = 1/(imMin[i] - imZT[i])/omega
            decimale = floor(log10(Cmax))
            indice = len(stdCapas)-1
            Ctot = 0
            Ccombi = []
            while True:
                while  stdCapas[indice]*10**decimale+Ctot > Cmax and indice>-1:
                    indice -= 1
                if indice == -1:
                    decimale -=1
                    indice =  len(stdCapas)-1
                elif stdCapas[indice]*10**decimale + Ctot  < Cmin:
                    Ctot += stdCapas[indice]*10**decimale
                    Ccombi.append(stdCapas[indice]*10**decimale)
                    decimale -= 1
                    indice =  len(stdCapas)-1
                else:
                    break
            Ctot += stdCapas[indice]*10**decimale
            Ccombi.append(stdCapas[indice]*10**decimale)
            Clump[i] = Ctot/1e-12
            imZ[i] = -1/(omega * Ctot)
            ZselfImg[i] = imZT[i] - imZ[i]
            Call.append(Ccombi)
        else:
            ZselfImg[i] = imZT[i]
            imZ[i] = 0
            Clump[i] = -float('inf')
            Call.append('None')
    return(Clump,imZ,ZselfImg,Call)


def Width(induct,L,bnd):
        opt = minimize_scalar(lambda w:abs(selfInducRect(L,w)*1e-7*omega-induct)**2,bounds=bnd,method='bounded')
        return(opt.x)


if not flagSym:
    x0 = zeros(2*N)
else:
    x0 = zeros((N+1))



bnds=[]
if not flagSym:
    for i in range(N):
        bnds.append((imMinLeg[i]+ZminCLeg,imMaxLeg[i]))
    for i in range(N):
        bnds.append((imMinRing[i]+ZminCRing,imMaxRing[i]))
else:
    for i in range(N//2+1):
        bnds.append((imMinLeg[i]+ZminCLeg,imMaxLeg[i]))
    for i in range(N//2):
        bnds.append((imMinRing[i]+ZminCRing,imMaxRing[i]))
    bnds.append((imMinRing[N-1]+ZminCRing,imMaxRing[N-1]))
        
for i,val in enumerate(bnds):
    if val[0]>0:
        x0[i] = (val[0]+val[1])/2.0
    else:
        x0[i] = 0.0

bestfun = 1000000
ite = 30
while ite>0:
    for i,val in enumerate(bnds):
        x0[i] = x0[i]-randn(1)*bestfun*100
        if x0[i] < val[0]:
            x0[i] = val[0]*0.99
        elif x0[i]> val[1]:
            x0[i] = val[1]*0.99
            
    opt = minimize(toOptimize, x0, method='L-BFGS-B',bounds = bnds)
    if opt.fun<bestfun:
        ite = 30
        best = opt
        bestfun = best.fun
        print(bestfun)
    if bestfun < 1e-7:
        break
    ite -= 1
    
#opt = basinhopping(toOptimize, x0)#,bounds = bnds)
print(best.x)
if  any(best.x>[b[1] for b in bnds]):
    raise NameError('Error optimized larger than bound')

if flagSym:
    imZlT, imZrT = symZ(best.x)
else:
    imZlT, imZrT = best.x[:N],best.x[N:]#symZ(opt.x)
    
M,v,g,zetaL,zetaR = INode(imZlT,imZrT)
# figure()
# subplot(2,1,1)

# plot(abs(M))
# plot(abs(Il1))
# subplot(2,1,2)
# plot(angle(M))
# plot(angle(Il1))
if ZminCLeg<0:
    (Clumpl,imZl,ZselfLegImg,ClumpCompoLeg) = Capas2(imZlT,imMinLeg,imMaxLeg)
    widthLeg = [Width(Ind*2.0,lengthLeg,(widthLegMin,widthLegMax)) for Ind in ZselfLegImg]
else:
    widthLeg = [Width(Ind*2.0,lengthLeg,(widthLegMin,widthLegMax)) for Ind in imZlT]

if ZminCRing<0:
    (Clumpr,imZr,ZselfRingImg,ClumpCompoRing) = Capas2(imZrT,imMinRing,imMaxRing)
    widthRing = [Width(Ind,lR,(widthRingMin,widthRingMax)) for (Ind,lR) in zip(ZselfRingImg,ringLengths)]
else:
    widthRing = [Width(Ind,lR,(widthRingMin,widthRingMax)) for (Ind,lR) in zip(imZlrT,ringLengths)]
#res,U,DDU = UDU(imZl,imZr,Icst)
Lumpl=[]
Lumpr=[]


tnu = linspace(100e6,500e6,10000)


# Compute the self and mutual reactance
def CalGammaTot(omega):
    LLeg0,LRing0 = ComputeLHalf(omega/3e8*coefk,None,None)
    RLegHalf,RRing = ComputeRoverHalf(omega/3e8*coefk)
    SelfIndLeg  =  array([selfInducRect(lengthLeg,w)*1e-7*omega/2.0 for w in widthLeg])
    SelfIndRing =  array([selfInducRect(l,w)*1e-7*omega for (l,w) in zip(ringLengths,widthRing)])
    imZlnu = [0 if lu==-float('inf') else -1/omega/lu/1e-12 for lu in Clumpl]
    imZrnu = [0 if lu==-float('inf') else -1/omega/lu/1e-12 for lu in Clumpr]
    Zleg = LLeg0*1e-7*1j*omega   + RLegHalf + 1j*diag(imZlnu+SelfIndLeg)
    Zring = LRing0*1e-7*1j*omega + RRing    + 1j*diag(imZrnu+SelfIndRing)
    return(Gamma(Zleg,Zring),Zleg,Zring)

def savecsv(name,xy,title=""):
    with open(name,'wt') as f:
        if title:
            f.write(title)
            f.write("\n")
        keys = [key for key in xy.keys()]
        for ikey,key in enumerate(keys):
            f.write('%s'%key)
            if ikey<len(xy)-1:
                f.write(',')
        f.write('\n')
        for i1 in range(len(xy[keys[0]])):
            for ikey,key in enumerate(keys):
                val = xy[key][i1]
                print(val)
                if isinstance(val,numpy.int64):
                    f.write('%d'%val)
                elif isinstance(val,numpy.float64):
                    f.write('%e'%val)
                elif isinstance(val,complex):
                    if imag(val)>0:
                        f.write('%e+i%e'%(real(val),imag(val)))
                    else:
                        f.write('%e-i%e'%(real(val),-imag(val)))
                        
                if ikey<len(xy)-1:
                    f.write(',')
            f.write('\n')
                
            


# Workout the eigen modes from the mutual reactance
f = fopen('Voltage_on_nodes','wt')

Gamma2,Zl,Zr = CalGammaTot(omega)
u,s,v = svd(Gamma2)
u = u[:,-1:]
figure()
U = Zl@Il1
subplot(2,1,1)
plot(np.abs(u)/np.max(abs(u)),'b',label='last mode')
plot(abs(U)/max(abs(U)),'r',label='target')
title('Voltage on nodes')
legend()
subplot(2,1,2)
il=inv(Zl)@u
plot(np.abs(il)/np.max(np.abs(il)),'b',label='last mode')
plot(abs(Il1)/max(abs(Il1)),'r',label='target')
title('Current on rings')
savecsv('Voltage_on_nodes.csv',{'Node':arange(N),'Eigen-mode':u[:,0],'Target':U},'Voltage on nodes')
savecsv('Current_on_legs.csv',{'Node':arange(N),'Eigen-mode':il[:,0],'Target':Il1},'Intensity on legs')
# Calculus with a single port excitation

indice = N//2
Ztot =inv(CalGammaTot(omega)[0])
Z11 = Ztot[indice,indice]
figure()
plot(real(diag(Ztot)))
plot(imag(diag(Ztot)))
ylabel('Port impedance')
xlabel('port number')
savecsv('Impedance.csv',{'Node':arange(N),'Z(Ohms)':diag(Ztot)},'Equivalent impedance on nodes')


I = zeros((N),dtype=complex)
I[indice] = 1/(50+Z11)
U = Ztot@I
Il = inv(Zl)@U
figure()
subplot(2,1,1)
plot(abs(Il)/max(abs(Il)),'r',label='single port')
plot(abs(Il1)/max(abs(Il1)),'b',label='target')
legend()
title('Current on rings')

savecsv('Single_port_excitation.csv',{'Node':arange(N),'Current on legs':Il,'Target':Il1},'Current on legs')




subplot(2,1,2)
plot(angle(Il),'r',label='single port')
plot(angle(Il1),'b',label='target')
title('Current phase on rings')
legend()

if 0:
    indice = N//2
    Z11 = zeros((len(tnu)),dtype=complex)
    for inu,nu in enumerate(tnu):
        sys.stdout.write('\rProgress %3d %%'%int(inu/len(tnu)*100.0))
        Ztot =inv(CalGammaTot(2*pi*nu)[0])
        Z11[inu] = Ztot[indice,indice]
    
     
    
    figure()
    subplot(2,1,1)
    plot(tnu,abs(Z11))
    subplot(2,1,2)
    plot(tnu,abs((Z11-50)/(Z11+50)))

np.savez('geometrie', Rshield=R0, R=Rleg,lengthLeg=lengthLeg,Nlegs=len(theta),widthLeg=widthLeg,widthRing=widthRing,angles=theta,CLeg=Clumpl,CRing=Clumpr)   

  
def export_array(file,values):
    for value in values[:-1]:
        file.write('%e;'%value)
    file.write('%e\n'%values[-1])

    
def export():
    with open('geo.txt','wt') as f:
        f.write('R0=%f\n'%R0)
        f.write('Rleg=%f\n'%Rleg)
        f.write('lengthLeg=%f\n'%lengthLeg)
        f.write('N=%d\n'%len(theta))
        f.write('widthLeg=')
        export_array(f,widthLeg)
        f.write('widthRing=')
        export_array(f,widthRing)
        f.write('angles=')
        export_array(f,theta)
        f.write('Cleg=')
        export_array(f,Clumpl)
        f.write('Cring=')
        export_array(f,Clumpr)
        
export()      
